window.Gleam.initSearch([{"doc":"formz","title":"formz","content":"# formz\n\n[![Package Version](https://img.shields.io/hexpm/v/formz)](https://hex.pm/packages/formz)\n[![Hex Docs](https://img.shields.io/badge/hex-docs-ffaff3)](https://hexdocs.pm/formz/)\n\nA Gleam library for parsing and generating accessible HTML forms.\n\nHTML forms rendered in the browser and the data they are parsed into are\nintrinsically linked. Treating the markup and the parsing as two separate\nproblems to solve is inconvenient and leads to bugs. This library aims\nto make that link explicit and easy to manage, while making it really easy\nto make accessible forms.\n\nNote: This library is not particularly well-suited for generating one-off\nforms, but is more intended for use in projects where you have a few forms to\nmanage, and would like to keep the form markup and parsing logic in sync.  It\ntakes some amount of effort to make an actual form generator with markup and\nstyles, and that might not be worth it for a one-off form.  That said, a simple\nform generator is provided if you aren&#39;t opinionated about your markup.\n\n```sh\ngleam add formz@0.1\n```\n\n## Creating a form\n\nA `formz` form is a list of fields and a decoder function.  You construct the\ndecoder function as fields are added:\n\n```gleam\nimport formz\nimport formz/field.{field}\nimport formz_string/definitions\n\npub fn make_form() {\n  use username &lt;- formz.require(field(&quot;username&quot;), definitions.text_field())\n  use password &lt;- formz.require(field(&quot;password&quot;), definitions.password_field())\n\n  formz.create_form(#(username, password))\n}\n```\n\n## Creating fields\n\nThere are two arguments to adding a field to a form (seen above):\n\n1. A [Field](https://hexdocs.pm/formz/formz/field.html), which holds specific,\n   unique details about the field, such as its name, label, help text, disabled\n   state, etc.\n2. A [Definition](https://hexdocs.pm/formz/formz/definition.html), which\n   says (A) how to generate the HTML input element for the field, and (B) how\n   to parse the data from the field. These definitions are reusable and can be\n   shared across fields, forms and projects.\n\n### Field details\n\n```gleam\n// name is required, the other details are optional\nfield(named: &quot;username&quot;)\n|&gt; field.set_label(&quot;Username&quot;)\n|&gt; field.set_help_text(&quot;Only alphanumeric characters are allowed.&quot;)\n```\n\n```gleam\nfield(named: &quot;userid&quot;) |&gt; field.make_hidden |&gt; field.set_raw_value(&quot;42&quot;)\n```\n\n### Field definition\n\n[Defintions](https://hexdocs.pm/formz/formz/definition.html) are the heavy\ncompared to the lightness of fields; they take a bit more work to make as they\nare intended to be more reusable.\n\nThe first role of a `Defintion` is to generate the HTML widget for the field.\nThis library is format-agnostic and you can generate HTML widgets as raw\nstrings, Lustre elements, Nakai nodes, something else, etc. There are\ncurrently three `formz` libraries that provide common field definitions for the\nmost common HTML formats.\n\n- [formz_string](https://hexdocs.pm/formz_string/)\n- [formz_nakai](https://hexdocs.pm/formz_nakai/)\n- [formz_lustre](https://hexdocs.pm/formz_lustre/) (untested in a browser, I&#39;ve only done this server side)\n\nThe second role  of a `Definition` is to parse the data from the field. There\nare a two parts to this, as how you parse a field&#39;s value depends on if it is\noptional or required.  Not all scenarios can be cookie-cutter placed into an\n`Option`. So you need to provide two parse functions, one for when a field is\nrequired, and a second for when it&#39;s optional.\n\n```gleam\n/// you won&#39;t often need to do this directly (I think??).  The idea is that\n/// there&#39;d be libs with the definitions you need.\n\nimport formz/definition.{Definition}\nimport formz/field\nimport formz/validation\nimport formz/widget\nimport lustre/attribute\nimport lustre/element\nimport lustre/element/html\n\nfn password_widget(\n  field: field.Field,\n  args: widget.Args,\n) -&gt; element.Element(msg) {\n  html.input([\n    attribute.type_(&quot;password&quot;),\n    attribute.name(field.name),\n    attribute.id(args.id),\n    attribute.attribute(&quot;aria-labelledby&quot;, field.label),\n  ])\n}\n\npub fn password_field() {\n  Definition(\n    widget: password_widget,\n    parse: validation.string,\n    optional_parse: fn(parse, str) {\n      case str {\n        &quot;&quot; -&gt; Ok(option.None)\n        _ -&gt; parse(str)\n      }\n    },\n    // We need to have a stub value for each parser. The stubs are used when\n    // building the decoder and parse functions for the form.\n    stub: &quot;&quot;,\n    optional_stub: option.None,\n  )\n}\n```\n\n\n\n## Generating HTML for a form\n\nGenerally speaking, the idea with a `formz` form is that you are not going\nto generate the HTML for each field individually, but rather, you&#39;d use\na function to loop through each field, generating semantic, accessible\nmarkup for each one.\n\nThe specifics of how you would do this are going\nto vary greatly for each project and its styling/markup needs.\n\nHowever, the three `formz_*` libraries mentioned above all provide a\nsimple form generator function that you can use as is, or as a starting\npoint for your own.  `formz` is BYOS, Bring Your Own Stylesheet, so the\nbuilt-in form generators come unstyled. If there is interest, I could add\na super simple CSS file to get the ball rolling and make the default\nforms easier to use out of the box.\n\nThat said, you can also create the form HTML yourself, directly for each field.\nThere&#39;s [an example](https://github.com/bentomas/formz/blob/main/formz_demo/src/formz_demo/examples/custom_output.gleam)\nin the demo project showing how to do this.\n\n### Generating form HTML using the `formz_string` library\n\nThe built-in form generators leave it as homework to add the form tags and\nsubmit buttons.\n\n```gleam\nimport formz_string/simple\n\npub fn show_form(form) -&gt; String {\n  &quot;&lt;form method=\\&quot;post\\&quot;&gt;&quot;\n  &lt;&gt; simple.generate(form)\n  &lt;&gt; &quot;&lt;p&gt;&lt;button type\\&quot;submit\\&quot;&gt;Submit&lt;/button&gt;&lt;/p&gt;&quot;\n  &lt;&gt; &quot;&lt;/form&gt;&quot;\n}\n```\n\n\n## Parsing form data\n\nYou can parse a `formz` form with a tuple of values and names, typically from\na POST request.  Here we parse in a `wisp` handler:\n\n```gleam\npub fn handle_form_submission(req: Request) -&gt; Response {\n  use formdata &lt;- wisp.require_form(req)\n\n  let result = make_form()\n  |&gt; formz.data(formdata.values)\n  |&gt; formz.parse\n\n  case result {\n    Ok(credentials) -&gt; {\n      let #(username, password) = credentials\n      wisp.ok()\n      |&gt; wisp.html_body(string_builder.from_string(&quot;Hello &quot;&lt;&gt;username&lt;&gt;&quot;!&quot;))\n    }\n    Error(form_with_errors) -&gt; {\n      show_form(form_with_errors)\n    }\n  }\n}\n```\n\nHowever, often you want to parse a form, and then... you know... act on that\ndata, and in doing so you might discover more errors for the form.  In this\nsituation you can use `parse_then_try`:\n\n```gleam\npub fn handle_form_submission(req: Request) -&gt; Response {\n  use formdata &lt;- wisp.require_form(req)\n\n  let result = make_form()\n  |&gt; formz.data(formdata.values)\n  |&gt; formz.parse_then_try(fn(form, credentials) {\n    case credentials {\n      #(&quot;admin&quot; as username, &quot;l33t&quot;) -&gt; Ok(username)\n      #(&quot;admin&quot;, _) -&gt;\n        form\n        |&gt; formz.update_field(&quot;password&quot;, field.set_error(_, &quot;Wrong password&quot;))\n        |&gt; Error\n      _ -&gt;\n        form\n        |&gt; formz.update_field(&quot;username&quot;, field.set_error(_, &quot;Unknown username&quot;))\n        |&gt; Error\n    }\n  })\n\n  case result {\n    Ok(username) -&gt; {\n      wisp.ok()\n      |&gt; wisp.html_body(string_builder.from_string(&quot;Hello &quot; &lt;&gt; username &lt;&gt; &quot;!&quot;))\n    }\n    Error(form_with_errors) -&gt; {\n      show_form(form_with_errors)\n    }\n  }\n}\n```\n\n## See it in action\n\nThere is a [demo wisp app](https://github.com/bentomas/formz/tree/main/formz_demo)\nshowing a few interactive examples of how `formz` works in the repo.\n","url":"index.html"},{"doc":"formz/definition","title":"Definition","content":"pub type Definition(widget, required, optional) {\n  Definition(\n    widget: widget,\n    parse: fn(String) -&gt; Result(required, String),\n    stub: required,\n    optional_parse: fn(\n      fn(String) -&gt; Result(required, String),\n      String,\n    ) -&gt;\n      Result(optional, String),\n    optional_stub: optional,\n  )\n}\n\nDefinition(\n  widget: widget,\n  parse: fn(String) -&gt; Result(required, String),\n  stub: required,\n  optional_parse: fn(\n    fn(String) -&gt; Result(required, String),\n    String,\n  ) -&gt;\n    Result(optional, String),\n  optional_stub: optional,\n)\n\nwidget\n&lt;p&gt;The widget is used by the form generator to construct the HTML for the\nfield. It’s up to the form generator to decide on the widget’s type and\nhow it’s used.&lt;/p&gt;\n\nparse\n&lt;p&gt;This parse function takes the raw string from the parsed POST data\nand converts it to a Gleam type.  This &lt;code&gt;parse&lt;/code&gt; is for when a value\nis required, so it should return an error if the field is\nconsidered empty.&lt;/p&gt;\n\nstub\n&lt;p&gt;The callbacks pattern for generating a form requires a stub\nvalue for each field, because the actual decode function is called\nstep by step as the fields are added to the form and &lt;code&gt;formz&lt;/code&gt; learns\nthe form’s details as it goes.  This stub value is purely used\nfor navigating the decode function, and just needs to match the type\nof the real value that can be parsed.&lt;/p&gt;\n\noptional_parse\n&lt;p&gt;If a field is marked as optional, this function is called, with the\nabove parse as an argument.  The idea is that this function will\ncall out to the parse function if the field is not empty, and\nthis should only handle the case where the raw input value is empty.\nThis function is necessary because not all fields should just be parsed\ninto an &lt;code&gt;Option&lt;/code&gt; when they aren’t provided.\nFor example, an optional text field might be an empty string,\nan optional checkbox might be &lt;code&gt;False&lt;/code&gt;, and an optional select might\nbe &lt;code&gt;option.None&lt;/code&gt;.&lt;/p&gt;\n\noptional_stub\n&lt;p&gt;stub for the &lt;code&gt;optional_parse&lt;/code&gt; return value&lt;/p&gt;\n\nSynonyms:\nformz/definition.Definition\nformz/definition Definition","url":"formz/definition.html#Definition"},{"doc":"formz/definition","title":"make_simple_optional_parse","content":"pub fn make_simple_optional_parse() -&gt; fn(\n  fn(String) -&gt; Result(a, String),\n  String,\n) -&gt; Result(Option(a), String)\n A convenience function to make the simple optional parse function where\n if a value isn&#39;t provided, just return `option.None`, otherwise call out\n to the parse function and put it&#39;s value in `option.Some`.\nSynonyms:\nformz/definition.make_simple_optional_parse\nformz/definition make_simple_optional_parse","url":"formz/definition.html#make_simple_optional_parse"},{"doc":"formz/definition","title":"set_widget","content":"pub fn set_widget(\n  definition: Definition(a, b, c),\n  widget: a,\n) -&gt; Definition(a, b, c)\n Replace the widget that this `Definition` uses for rendering the field.  Most\n HTML inputs can be interchangeable, they all generate a `String` after all,\n but not all are the best UX.  This allows you to choose the one that is the\n most appropriate for your field.  However, the details of how this works\n are up to the form generator.\nSynonyms:\nformz/definition.set_widget\nformz/definition set_widget","url":"formz/definition.html#set_widget"},{"doc":"formz/definition","title":"validate","content":"pub fn validate(\n  def: Definition(a, b, c),\n  fun: fn(b) -&gt; Result(b, String),\n) -&gt; Definition(a, b, c)\n Chain additional validation onto the `parse` function.  This is\n useful if you don&#39;t need to change the returned type, but might have\n additional constraints.  Like say, requiring a `String` to be at least\n a certain length, or that an Int must be positive.\n\n ### Example\n \n field\n   |&gt; validate(fn(i) {\n     case i &gt; 0 {\n       True -&gt; Ok(i)\n       False -&gt; Error(&quot;must be positive&quot;)\n     }\n   }),\n \nSynonyms:\nformz/definition.validate\nformz/definition validate","url":"formz/definition.html#validate"},{"doc":"formz/definition","title":"formz/definition","content":" A `Definition` is the second argument needed to add a field to a form. It\n is what describes how a field works, e.g. how it looks and how it&#39;s parsed.\n It is the heavy compared to the lightness of a [Field](https://hexdocs.pm/formz/formz/field.html);\n they take a bit more work to make as they are intended to be reusable.\n\n The first role of a `Defintion` is to generate the HTML widget for the field.\n This library is format-agnostic and you can generate HTML widgets as raw\n strings, Lustre elements, Nakai nodes, something else, etc. There are\n currently three `formz` libraries that provide common widgets (and field\n definitions) for the most common HTML formats.\n\n - [formz_string](https://hexdocs.pm/formz_string/)\n - [formz_nakai](https://hexdocs.pm/formz_nakai/)\n - [formz_lustre](https://hexdocs.pm/formz_lustre/) (untested in a browser,\n   would it be useful there??)\n\n The second role  of a `Definition` is to parse the data from the field.\n There are a two parts to this, as how you parse a field&#39;s value depends on\n if it is optional or required.  Not all scenarios can be cookie-cutter\n placed into an `Option`. So you need to provide two parse functions, one\n for when a field is required, and a second for when it&#39;s optional.\n\n ### Example password field definition\n\n ```gleam\n /// you won&#39;t often need to do this directly (I think??).  The idea is that\n /// there&#39;d be libs with the definitions you need.\n\n import formz/definition.{Definition}\n import formz/field\n import formz/validation\n import formz/widget\n import lustre/attribute\n import lustre/element\n import lustre/element/html\n\n fn password_widget(\n   field: field.Field,\n   args: widget.Args,\n ) -&gt; element.Element(msg) {\n   html.input([\n     attribute.type_(&quot;password&quot;),\n     attribute.name(field.name),\n     attribute.id(args.id),\n     attribute.attribute(&quot;aria-labelledby&quot;, field.label),\n   ])\n }\n\n pub fn password_field() {\n   Definition(\n     widget: password_widget,\n     parse: validation.string,\n     optional_parse: fn(parse, str) {\n       case str {\n         &quot;&quot; -&gt; Ok(option.None)\n         _ -&gt; parse(str)\n       }\n     },\n     // We need to have a stub value for each parser. The stubs are used when\n     // building the decoder and parse functions for the form.\n     stub: &quot;&quot;,\n     optional_stub: option.None,\n   )\n }\n ```","url":"formz/definition.html"},{"doc":"formz/field","title":"Field","content":"pub type Field {\n  Valid(\n    name: String,\n    label: String,\n    help_text: String,\n    disabled: Bool,\n    required: Bool,\n    hidden: Bool,\n    value: String,\n  )\n  Invalid(\n    name: String,\n    label: String,\n    help_text: String,\n    disabled: Bool,\n    required: Bool,\n    hidden: Bool,\n    value: String,\n    error: String,\n  )\n}\n\nValid(\n  name: String,\n  label: String,\n  help_text: String,\n  disabled: Bool,\n  required: Bool,\n  hidden: Bool,\n  value: String,\n)\n\nname\n&lt;p&gt;The name of the field. The only truly required information for a field.\nThis is used to identify the field in the form. It should be unique for\neach form, and is untested with any values other than strings solely\nconsisting of alphanumeric characters and underscores.&lt;/p&gt;\n\nlabel\n&lt;p&gt;This library thinks of a label as required, but will make one for you from\nthe name if you don’t provide one via the &lt;code&gt;field&lt;/code&gt; function. For\naccessibility reasons, a field should always provide a label and all\nthe maintained form generators will output one.&lt;/p&gt;\n\nhelp_text\n&lt;p&gt;Optional help text for the field. This is used to provide additional\ninstructions or context for the field.  It is up to the form generator\nto decide if and how to display this text.&lt;/p&gt;\n\ndisabled\n&lt;p&gt;Whether the field is disabled. A disabled field is not editable in\nthe browser.  However, there is nothing stopping a user from changing\nthe value or submitting a different value via other means, so (presently)\nthis doesn’t mean the value cannot be tampered with.&lt;/p&gt;\n\nrequired\n&lt;p&gt;Whether the field is required. This field is not functional, but is purely\nfor whether or not the form generator should indicate to the user that the\nfield is required.  Add the field to the form with either &lt;code&gt;optional&lt;/code&gt; or\n&lt;code&gt;required&lt;/code&gt; methods to control this functionally.  Those methods will make\nsure this field is set correctly.&lt;/p&gt;\n\nhidden\n&lt;p&gt;Whether the field is hidden. A hidden field is not displayed in the browser.&lt;/p&gt;\n\nvalue\n&lt;p&gt;The value of the field. This is normally set from via the &lt;code&gt;data&lt;/code&gt; function on\nthe form, but this can be set manually if you need default or initial values\nfor a particular field.&lt;/p&gt;\n\nInvalid(\n  name: String,\n  label: String,\n  help_text: String,\n  disabled: Bool,\n  required: Bool,\n  hidden: Bool,\n  value: String,\n  error: String,\n)\n\nerror\n&lt;p&gt;An error message for the field. This is set if a parser function from a\ndefinition fails for a field.  You can also set it yourself if you have\nan error that isn’t related to parsing, but is with the actual data\nitself.&lt;/p&gt;\n\nSynonyms:\nformz/field.Field\nformz/field Field","url":"formz/field.html#Field"},{"doc":"formz/field","title":"field","content":"pub fn field(named name: String) -&gt; Field\n Create a field with the given name. It uses [justin.sentence_case](https://hexdocs.pm/justin/justin.html#sentence_case)\n to create a label. You can override the label with the `set_label` function.\n\n \n field(&quot;name&quot;)\n |&gt; set_label(&quot;Full Name&quot;)\n \nSynonyms:\nformz/field.field\nformz/field field","url":"formz/field.html#field"},{"doc":"formz/field","title":"make_disabled","content":"pub fn make_disabled(field: Field) -&gt; Field\n\nSynonyms:\nformz/field.make_disabled\nformz/field make_disabled","url":"formz/field.html#make_disabled"},{"doc":"formz/field","title":"make_hidden","content":"pub fn make_hidden(field: Field) -&gt; Field\n\nSynonyms:\nformz/field.make_hidden\nformz/field make_hidden","url":"formz/field.html#make_hidden"},{"doc":"formz/field","title":"set_disabled","content":"pub fn set_disabled(field: Field, disabled: Bool) -&gt; Field\n\nSynonyms:\nformz/field.set_disabled\nformz/field set_disabled","url":"formz/field.html#set_disabled"},{"doc":"formz/field","title":"set_error","content":"pub fn set_error(field: Field, error: String) -&gt; Field\n\nSynonyms:\nformz/field.set_error\nformz/field set_error","url":"formz/field.html#set_error"},{"doc":"formz/field","title":"set_help_text","content":"pub fn set_help_text(field: Field, help_text: String) -&gt; Field\n\nSynonyms:\nformz/field.set_help_text\nformz/field set_help_text","url":"formz/field.html#set_help_text"},{"doc":"formz/field","title":"set_hidden","content":"pub fn set_hidden(field: Field, hidden: Bool) -&gt; Field\n\nSynonyms:\nformz/field.set_hidden\nformz/field set_hidden","url":"formz/field.html#set_hidden"},{"doc":"formz/field","title":"set_label","content":"pub fn set_label(field: Field, label: String) -&gt; Field\n\nSynonyms:\nformz/field.set_label\nformz/field set_label","url":"formz/field.html#set_label"},{"doc":"formz/field","title":"set_name","content":"pub fn set_name(field: Field, name: String) -&gt; Field\n\nSynonyms:\nformz/field.set_name\nformz/field set_name","url":"formz/field.html#set_name"},{"doc":"formz/field","title":"set_raw_value","content":"pub fn set_raw_value(field: Field, value: String) -&gt; Field\n\nSynonyms:\nformz/field.set_raw_value\nformz/field set_raw_value","url":"formz/field.html#set_raw_value"},{"doc":"formz/field","title":"formz/field","content":" A `Field` is the first argument needed to add a field to a form. It contains\n information about this specific field, like it&#39;s name, label, or (optional)\n help_text. There is a convenience function to create a field with just a name,\n and then you can use the rest of the functions to set just the values you\n need to change.\n\n ```gleam\n field(&quot;name&quot;) |&gt; set_label(&quot;Full Name&quot;)\n ```\n\n ```gleam\n field(&quot;name&quot;)\n |&gt; set_label(&quot;Full Name&quot;)\n |&gt; set_help_text(&quot;You can make one up if you&#39;d like.&quot;)\n ```","url":"formz/field.html"},{"doc":"formz/subform","title":"SubForm","content":"pub type SubForm {\n  SubForm(name: String, label: String, help_text: String)\n}\n\nSubForm(name: String, label: String, help_text: String)\n\nname\n&lt;p&gt;The name of the subform. This is used to prefix all the fields of the\nsubform, so it should be unique for each subform added to aform.\nIt is untested with any values other than strings consisting solely\nof alphanumeric characters and underscores.&lt;/p&gt;\n\nlabel\n&lt;p&gt;The label of the subform. This is completely optional, but if the\nsubform is rendered inside a &lt;code&gt;&amp;lt;fieldset&amp;gt;&lt;/code&gt; then it is &lt;a href=&quot;https://www.w3.org/WAI/tutorials/forms/grouping/&quot;&gt;recommended&lt;/a&gt;\nto have a &lt;code&gt;&amp;lt;legend&amp;gt;&lt;/code&gt; with this label.&lt;/p&gt;\n\nhelp_text\n&lt;p&gt;Help text for the subform.  There is less of a standard for this, but\nagain, if rendered in a &lt;code&gt;&amp;lt;fieldset&amp;gt;&lt;/code&gt; then &lt;code&gt;area-describedby&lt;/code&gt; can be used\nto point to an element with this help text.&lt;/p&gt;\n\nSynonyms:\nformz/subform.SubForm\nformz/subform SubForm","url":"formz/subform.html#SubForm"},{"doc":"formz/subform","title":"set_help_text","content":"pub fn set_help_text(sub: SubForm, help_text: String) -&gt; SubForm\n\nSynonyms:\nformz/subform.set_help_text\nformz/subform set_help_text","url":"formz/subform.html#set_help_text"},{"doc":"formz/subform","title":"set_label","content":"pub fn set_label(sub: SubForm, label: String) -&gt; SubForm\n\nSynonyms:\nformz/subform.set_label\nformz/subform set_label","url":"formz/subform.html#set_label"},{"doc":"formz/subform","title":"set_name","content":"pub fn set_name(sub: SubForm, name: String) -&gt; SubForm\n\nSynonyms:\nformz/subform.set_name\nformz/subform set_name","url":"formz/subform.html#set_name"},{"doc":"formz/subform","title":"subform","content":"pub fn subform(name: String) -&gt; SubForm\n Create a subform with the given name. It uses [justin.sentence_case](https://hexdocs.pm/justin/justin.html#sentence_case)\n to create a label. You can override the label with the `set_label` function.\n\n \n subform(&quot;address&quot;)\n |&gt; set_label(&quot;Shipping Address&quot;)\n \nSynonyms:\nformz/subform.subform\nformz/subform subform","url":"formz/subform.html#subform"},{"doc":"formz/subform","title":"formz/subform","content":" Detials about a subform being added to a form. There is a convenience\n function to create a field with just a name, and then you can use the rest\n of the functions to set just the values you need to change.","url":"formz/subform.html"},{"doc":"formz","title":"Form","content":"pub opaque type Form(widget, output)\n The `widget` type is the set by the `Definition`s used to add fields for\n this form, and has the details of how to turn given fields into HTML inputs.\n\n The `output` type is the type of the decoded data from the form. This is\n set by the `create_form` function, after all the fields have been added.\n\nSynonyms:\nformz.Form\nformz Form","url":"formz.html#Form"},{"doc":"formz","title":"FormItem","content":"pub type FormItem(widget) {\n  Field(field.Field, widget: widget)\n  SubForm(subform.SubForm, items: List(FormItem(widget)))\n}\n A form contains a list of fields and subforms. You primarily only use these\n when writing a form generator function.  You can also manipulate these\n after a form has been created, to change things like labels, help text, etc.\n There are specific functions,`update_field` and `update_subform`, to help\n with this, so you don&#39;t have to pattern match when updating a specific item.\n\n \n let form = make_form()\n form.update_field(&quot;name&quot;, field.set_label(_, &quot;Full Name&quot;))\n \nField(field.Field, widget: widget)\n\n\nSubForm(subform.SubForm, items: List(FormItem(widget)))\n\n\nSynonyms:\nformz.FormItem\nformz FormItem","url":"formz.html#FormItem"},{"doc":"formz","title":"create_form","content":"pub fn create_form(thing: a) -&gt; Form(b, a)\n Create an empty form that only parses to `thing`. This is primarily\n intended to be the final return value of a chain of callbacks that adds\n the form&#39;s fields.\n\n \n create_form(1)\n |&gt; parse\n # -&gt; Ok(1)\n \n \n fn make_form() {\n   use field1 &lt;- formz.require(field(&quot;field1&quot;), definitions.text_field())\n   use field2 &lt;- formz.require(field(&quot;field2&quot;), definitions.text_field())\n   use field3 &lt;- formz.require(field(&quot;field3&quot;), definitions.text_field())\n\n   create_form(#(field1, field2, field3))\n }\n \nSynonyms:\nformz.create_form\nformz create_form","url":"formz.html#create_form"},{"doc":"formz","title":"data","content":"pub fn data(\n  form: Form(a, b),\n  input_data: List(#(String, String)),\n) -&gt; Form(a, b)\n Add input data to this form. This will set the raw string value of the fields.\n It does not trigger any parsing, so you can also use this to set default values\n (if you do it in your form generator function) or initial values (if you do it\n before rendering an empty form).\n\n The input data is a list of tuples, where the first element is the name of the\n field and the second element is the value to set.  If the field does not exist\n the data is ignored, and if multiple values are given for the same field, the\n last one wins.\nSynonyms:\nformz.data\nformz data","url":"formz.html#data"},{"doc":"formz","title":"get","content":"pub fn get(\n  form: Form(a, b),\n  name: String,\n) -&gt; Result(FormItem(a), Nil)\n Get the [`FormItem`](https://hexdocs.pm/formz/formz.html#FormItem) with the\n given name.  If multiple items have the same name, the first one is returned.\nSynonyms:\nformz.get\nformz get","url":"formz.html#get"},{"doc":"formz","title":"items","content":"pub fn items(form: Form(a, b)) -&gt; List(FormItem(a))\n Get each [`FormItem`](https://hexdocs.pm/formz/formz.html#FormItem) added\n to the form.  Any time a field or subform are added, a FormItem is created.\nSynonyms:\nformz.items\nformz items","url":"formz.html#items"},{"doc":"formz","title":"optional","content":"pub fn optional(\n  field: Field,\n  is definition: Definition(a, b, c),\n  rest fun: fn(c) -&gt; Form(a, d),\n) -&gt; Form(a, d)\n Add an optional field to a form.\n\n This will use both the `parse` and `optional_parse` functions from the\n definition to parse the input data when parsing this field.  Ultimately\n whether a field is actually optional or not comes down to the details\n of the definition.\n\n The final argument is a callback that will be called when the form\n is being constructed to look for more fields; validated to check for errors;\n and when the form is being parsed, to decode the input data.  **For this\n reason, the callback should be a function without side effects.** It can be\n called any number of times. Don&#39;t do anything but create the type with the\n data you need!\nSynonyms:\nformz.optional\nformz optional","url":"formz.html#optional"},{"doc":"formz","title":"parse","content":"pub fn parse(form: Form(a, b)) -&gt; Result(b, Form(a, b))\n Parse the form.  This means step through the fields one by one, parsing\n them individually.  If any field fails to parse, the whole form is considered\n invalid, however it will still continue parsing the rest of the fields to\n collect all errors.  This is useful for showing all errors at once.  If no\n fields fail to parse, the decoded value is returned, which is the value given\n to `create_form`.\n\n If you&#39;d like to parse the form but not get the output, so you can give\n feedback to a user in response to input, you can use `validate` or `validate_all`.\nSynonyms:\nformz.parse\nformz parse","url":"formz.html#parse"},{"doc":"formz","title":"parse_then_try","content":"pub fn parse_then_try(\n  form: Form(a, b),\n  apply fun: fn(Form(a, b), b) -&gt; Result(c, Form(a, b)),\n) -&gt; Result(c, Form(a, b))\n Parse the form, then apply a function to the output if it was successful.\n This is a very thin wrapper around `parse` and `result.try`, but the\n difference being it will pass the form along to the function as a second\n argument in addition to the successful result.  This allows you to easily\n update the form fields with errors or other information based on the output.\n\n This is useful for situations where you can have errors in the form that\n aren&#39;t easily checked in simple parsing functions.  Like, say, hitting a\n db to check if a username is taken.\n\n \n make_form()\n |&gt; data(form_data)\n |&gt; parse_then_try(fn(username, form) {\n   case is_username_taken(username) {\n     Ok(false) -&gt; Ok(form)\n     Ok(true) -&gt; update_field(form, &quot;username&quot;, field.set_error(_, &quot;Username is taken&quot;))\n   }\n }\nSynonyms:\nformz.parse_then_try\nformz parse_then_try","url":"formz.html#parse_then_try"},{"doc":"formz","title":"require","content":"pub fn require(\n  field: Field,\n  is definition: Definition(a, b, c),\n  rest fun: fn(b) -&gt; Form(a, d),\n) -&gt; Form(a, d)\n Add a required field to a form.\n\n This will use only the `parse` function from the definition to parse the\n input data when parsing this field. Ultimately whether a field is actually\n required or not comes down to the details of the definition.\n\n This will also set the `required` value on the field to `True`.  Form\n generators can use this to mark the HTML input elements as required for\n accessibility.\n\n The final argument is a callback that will be called when the form\n is being constructed to look for more fields; validated to check for errors;\n and when the form is being parsed, to decode the input data.  **For this\n reason, the callback should be a function without side effects.** It can be\n called any number of times. Don&#39;t do anything but create the type with the\n data you need!\nSynonyms:\nformz.require\nformz require","url":"formz.html#require"},{"doc":"formz","title":"subform","content":"pub fn subform(\n  details: SubForm,\n  sub: Form(a, b),\n  fun: fn(b) -&gt; Form(a, c),\n) -&gt; Form(a, c)\n Add a form as a subform.  This will essentially append the fields from the\n subform to the current form, prefixing their names with the name of the\n subform.  Form generators will still see the fields as a set though, so they\n can be marked up as a group for accessibility reasons.\n\n The final argument is a callback that will be called when the form\n is being constructed to look for more fields; validated to check for errors;\n and when the form is being parsed, to decode the input data.  **For this\n reason, the callback should be a function without side effects.** It can be\n called any number of times. Don&#39;t do anything but create the type with the\n data you need!\nSynonyms:\nformz.subform\nformz subform","url":"formz.html#subform"},{"doc":"formz","title":"update","content":"pub fn update(\n  form: Form(a, b),\n  name: String,\n  fun: fn(FormItem(a)) -&gt; FormItem(a),\n) -&gt; Form(a, b)\n Update the [`FormItem`](https://hexdocs.pm/formz/formz.html#FormItem) with\n the given name using the provided function.  If multiple items have the same\n name, it will be called on all of them.\nSynonyms:\nformz.update\nformz update","url":"formz.html#update"},{"doc":"formz","title":"update_field","content":"pub fn update_field(\n  form: Form(a, b),\n  name: String,\n  fun: fn(Field) -&gt; Field,\n) -&gt; Form(a, b)\n Update the [`Field`](https://hexdocs.pm/formz/formz/field.html) with\n the given name using the provided function.  If multiple items have the same\n name, it will be called on all of them.\n\n \n let form = make_form()\n update(form, &quot;name&quot;, field.set_label(_, &quot;Full Name&quot;))\n \nSynonyms:\nformz.update_field\nformz update_field","url":"formz.html#update_field"},{"doc":"formz","title":"update_subform","content":"pub fn update_subform(\n  form: Form(a, b),\n  name: String,\n  fun: fn(SubForm) -&gt; SubForm,\n) -&gt; Form(a, b)\n Update the [`SubForm`](https://hexdocs.pm/formz/formz/subform.html) with\n the given name using the provided function.  If multiple subforms have the same\n name, it will be called on all of them.\n\n \n let form = make_form()\n update(form, &quot;name&quot;, subform.set_help_text(_, &quot;...&quot;))\n \nSynonyms:\nformz.update_subform\nformz update_subform","url":"formz.html#update_subform"},{"doc":"formz","title":"validate","content":"pub fn validate(\n  form: Form(a, b),\n  names: List(String),\n) -&gt; Form(a, b)\n Validate specific fields of the form.  This is similar to `parse`, but\n instead of returning the decoded output if there are no errors, it returns\n the valid form.  This is useful for if you want to be able to give feedback\n to the user about whether certain fields are valid or not. In this case you\n could just validate only fields that the user has interacted with.\nSynonyms:\nformz.validate\nformz validate","url":"formz.html#validate"},{"doc":"formz","title":"validate_all","content":"pub fn validate_all(form: Form(a, b)) -&gt; Form(a, b)\n Validate all the fields in the form.  This is similar to `parse`, but\n instead of returning the decoded output if there are no errors, it returns\n the valid form.  This is useful for if you want to be able to give feedback\n to the user about whether certain fields are valid or not.\nSynonyms:\nformz.validate_all\nformz validate_all","url":"formz.html#validate_all"},{"doc":"formz","title":"formz","content":" A form is a list of fields and a decoder function. This module uses a\n series of callbacks to construct a decoder function as the fields are\n being added to the form.  The idea is that you&#39;d have a function that\n makes the form using the `use` syntax, and then be able to use the form\n later for parsing or rendering in different contexts.\n\n ### Examples\n\n ```gleam\n fn make_form() {\n  use name &lt;- formz.require(field(&quot;name&quot;), defintions.text_field())\n\n  formz.create_form(name)\n }\n\n fn process_form() {\n   make_form()\n   |&gt; formz.data([#(&quot;name&quot;, &quot;Louis&quot;))])\n   |&gt; formz.parse\n   # -&gt; Ok(&quot;Louis&quot;)\n }\n ```\n\n ```gleam\n fn make_form() {\n  use greeting &lt;- optional(field(&quot;greeting&quot;), defintions.text_field())\n  use name &lt;- optional(field(&quot;name&quot;), defintions.text_field())\n\n  formz.create_form(greeting &lt;&gt; &quot; &quot; &lt;&gt; name)\n }\n\n fn process_form() {\n   make_form()\n   |&gt; data([#(&quot;greeting&quot;, &quot;Hello&quot;), #(&quot;name&quot;, &quot;World&quot;)])\n   |&gt; formz.parse\n   # -&gt; Ok(&quot;Hello World&quot;)\n }\n ```","url":"formz.html"},{"doc":"formz/validation","title":"and","content":"pub fn and(\n  previous: fn(a) -&gt; Result(b, String),\n  next: fn(b) -&gt; Result(c, String),\n) -&gt; fn(a) -&gt; Result(c, String)\n Chain validations together.\n\n ## Examples\n\n \n let is_even = fn(num) {\n   case num % 2 == 0 {\n     True -&gt; Ok(num)\n     False -&gt; Error(&quot;must be even&quot;)\n   }\n }\n let check = and(int, is_even)\n\n check(&quot;2&quot;)\n # -&gt; Ok(2)\n\n check(&quot;hi&quot;)\n # -&gt; Error(&quot;must be a whole number&quot;)\n\n check(&quot;1&quot;)\n # -&gt; Error(&quot;must be even&quot;)\nSynonyms:\nformz/validation.and\nformz/validation and","url":"formz/validation.html#and"},{"doc":"formz/validation","title":"email","content":"pub fn email(input: String) -&gt; Result(String, String)\n Parse the input as a String that looks like an email address, i.e. it\n contains an `@` character, with at least one other character on either\n side\n\n (this behavior more closely matches what the browser does than just\n checking for an `@`).\n\n ## Examples\n\n \n email(&quot;hello@example.com&quot;)\n # -&gt; Ok(&quot;hello@example.com&quot;)\n \n \n email(&quot;@&quot;)\n # -&gt; Error(&quot;Must be an email address&quot;)\n \n \n email(&quot;hello&quot;)\n # -&gt; Error(&quot;Must be an email address&quot;)\n \nSynonyms:\nformz/validation.email\nformz/validation email","url":"formz/validation.html#email"},{"doc":"formz/validation","title":"int","content":"pub fn int(str: String) -&gt; Result(Int, String)\n Parse the input as an int.\n\n ## Examples\n\n \n int(&quot;1&quot;)\n # -&gt; Ok(1)\n \n\n \n int(&quot;3.4&quot;)\n # -&gt; Error(&quot;Must be a whole number&quot;)\n \n \n int(&quot;hello&quot;)\n # -&gt; Error(&quot;Must be a whole number&quot;)\n \nSynonyms:\nformz/validation.int\nformz/validation int","url":"formz/validation.html#int"},{"doc":"formz/validation","title":"list_item_by_index","content":"pub fn list_item_by_index(\n  variants: List(a),\n) -&gt; fn(String) -&gt; Result(a, String)\n Validates that the input is one from a list of allowed values. Takes a list\n of Gleam values that can be chosen.  This uses the index of the item in to\n find the desired value.\n\n ## Examples\n\n \n enum([&quot;One&quot;,&quot;Two&quot;,&quot;Three&quot;])(&quot;1&quot;)\n # -&gt; Ok(&quot;Two&quot;)\n \n\n \n enum([True, False])(&quot;42&quot;)\n # -&gt; Error(&quot;must be an item in list&quot;)\n \n\n \n enum([True, False])(&quot;ok&quot;)\n # -&gt; Error(&quot;must be an item in list&quot;)\n \nSynonyms:\nformz/validation.list_item_by_index\nformz/validation list_item_by_index","url":"formz/validation.html#list_item_by_index"},{"doc":"formz/validation","title":"non_empty_string","content":"pub fn non_empty_string(str: String) -&gt; Result(String, String)\n Trim and leave the input as it is, but verify it is non-empty.\n\n \n non_empty_string(&quot;hello&quot;)\n # -&gt; Ok(&quot;hello&quot;)\n \n\n \n non_empty_string(&quot;  &quot;)\n # -&gt; Error(&quot;is required&quot;)\n \nSynonyms:\nformz/validation.non_empty_string\nformz/validation non_empty_string","url":"formz/validation.html#non_empty_string"},{"doc":"formz/validation","title":"number","content":"pub fn number(str: String) -&gt; Result(Float, String)\n Parse the input as a float.  this is forgiving and will also parse\n ints into floats.\n\n ## Examples\n\n \n number(&quot;1&quot;)\n # -&gt; Ok(1.0)\n \n\n \n number(&quot;3.4&quot;)\n # -&gt; Ok(3.4)\n \n \n number(&quot;hello&quot;)\n # -&gt; Error(&quot;Must be a number&quot;)\n \nSynonyms:\nformz/validation.number\nformz/validation number","url":"formz/validation.html#number"},{"doc":"formz/validation","title":"on","content":"pub fn on(val: String) -&gt; Result(Bool, String)\n Parse the input as a boolean, where only &quot;on&quot; is True and allowed.\n All other values are an error.  This is useful for HTML checkboxes, which\n the browser sends the empty string if unchecked, and `&quot;on&quot;` if  checked.\n\n ## Examples\n\n \n on(&quot;on&quot;)\n # -&gt; Ok(True)\n \n\n \n on(&quot;&quot;)\n # -&gt; Error(&quot;is required&quot;)\n \n\n \n on(&quot;hi&quot;)\n # -&gt; Error(&quot;is required&quot;)\n \nSynonyms:\nformz/validation.on\nformz/validation on","url":"formz/validation.html#on"},{"doc":"formz/validation","title":"replace_error","content":"pub fn replace_error(\n  previous: fn(a) -&gt; Result(b, String),\n  error: String,\n) -&gt; fn(a) -&gt; Result(b, String)\n Replace the error message of a validation with a new one.  Most of the built-in\n error messages are pretty rudimentary.\nSynonyms:\nformz/validation.replace_error\nformz/validation replace_error","url":"formz/validation.html#replace_error"},{"doc":"formz/validation","title":"formz/validation","content":"","url":"formz/validation.html"},{"doc":"formz/widget","title":"Args","content":"pub type Args {\n  Args(\n    id: String,\n    labelled_by: LabelledBy,\n    described_by: DescribedBy,\n  )\n}\n\nArgs(\n  id: String,\n  labelled_by: LabelledBy,\n  described_by: DescribedBy,\n)\n\nid\n&lt;p&gt;The id of the input element.&lt;/p&gt;\n\nlabelled_by\n&lt;p&gt;Details of how the input is labelled. Some sort of label is required for accessibility.&lt;/p&gt;\n\ndescribed_by\n&lt;p&gt;Details of how the input is described. This is optional, but can be useful for accessibility.&lt;/p&gt;\n\nSynonyms:\nformz/widget.Args\nformz/widget Args","url":"formz/widget.html#Args"},{"doc":"formz/widget","title":"DescribedBy","content":"pub type DescribedBy {\n  DescribedByElementsWithIds(ids: List(String))\n  DescribedByNone\n}\n\nDescribedByElementsWithIds(ids: List(String))\n The input is described by elements with the specified ids. This is useful\n for additional instructions or error messages.\n\nDescribedByNone\n\n\nSynonyms:\nformz/widget.DescribedBy\nformz/widget DescribedBy","url":"formz/widget.html#DescribedBy"},{"doc":"formz/widget","title":"LabelledBy","content":"pub type LabelledBy {\n  LabelledByLabelFor\n  LabelledByFieldValue\n  LabelledByElementsWithIds(ids: List(String))\n}\n\nLabelledByLabelFor\n The input is labelled by a `&lt;label&gt;` element with a `for` attribute\n pointing to this input&#39;s id. This has the best accessibility support\n and should be [preferred when possible](https://www.w3.org/WAI/tutorials/forms/labels/).\n\nLabelledByFieldValue\n The input should be labelled using the `Field`&#39;s `label` field.\n\nLabelledByElementsWithIds(ids: List(String))\n The input is labelled by elements with the specified ids.\n\nSynonyms:\nformz/widget.LabelledBy\nformz/widget LabelledBy","url":"formz/widget.html#LabelledBy"},{"doc":"formz/widget","title":"Widget","content":"pub type Widget(format) =\n  fn(field.Field, Args) -&gt; format\n\n\nSynonyms:\nformz/widget.Widget\nformz/widget Widget","url":"formz/widget.html#Widget"},{"doc":"formz/widget","title":"formz/widget","content":" The goal of a &quot;widget&quot; in `formz` is to produce an HTML input like\n `&lt;input&gt;`, `&lt;select&gt;`, or `&lt;textarea&gt;`. In a [`Definition`](https://hexdocs.pm/formz/formz/definition.html),\n a widget can be any Gleam type, and it&#39;s up to the form generator being\n used to know the exact type you need.\n\n That said, in the bundled form generators a widget is a function that\n takes the details of a field and some render time arguments that the form\n generator needs to construct an input.  This module is for those form\n generators, and it&#39;s use is optional if you have different needs.","url":"formz/widget.html"}]);